﻿using System;
using System.Threading;

namespace producerConsumer
{
    /// <summary>
    /// BoundedBuffer has a limited capacity.<br />
    /// Trying to Put an element into a full buffer will make the calling thread wait.
    /// </summary>
    public class BoundedBuffer : AbstractBuffer
    {
        private readonly int _capacity;
        private Boolean _ending = false;

        /// <summary>
        /// Constructs a new BoundedBuffer object
        /// </summary>
        /// <param name="capacity">The capacity of the BoundedBuffer object</param>
        /// <exception cref="ArgumentOutOfRangeException">If the capacity is less than or equal to zero</exception>
        public BoundedBuffer(int capacity)
        {
            if (capacity <= 0)
            {
                throw new ArgumentOutOfRangeException("Negative size: " + capacity);
            }
            this._capacity = capacity;
        }

        public int Capacity
        {
            get { return _capacity; }
        }

        /// <summary>
        /// Returns true if the buffer is full, false otherwise
        /// </summary>
        /// <returns>true if the buffer is full, false otherwise</returns>
        public Boolean IsFull()
        {
            Monitor.Enter(Queue);
            try
            {
                return (Queue.Count == _capacity);
            }
            finally
            {
                Monitor.Exit(Queue);
            }
        }

        /// <summary>
        /// Puts an element into the end of this buffer.<br />
        /// If the buffer is full, the calling thread will wait until another thread has taken an element from the buffer
        /// </summary>
        /// <param name="element">new element to be put into this buffer</param>
        public override void Put(int element)
        {
            if (_ending)
            {
                throw new InvalidOperationException("Trying to put after putting the STOP_VALUE");
            }
            Monitor.Enter(Queue);
            try
            {
                while (IsFull())
                {
                    Monitor.Wait(Queue);
                }
                Queue.Enqueue(element);
                if (Queue.Count == 1)
                {
                    Monitor.PulseAll(Queue);
                }
            }
            finally
            {
                Monitor.Exit(Queue);
            }
        }

        public override int Take()
        {
            Monitor.Enter(Queue);
            try
            {
                if (_ending) { return Producer.LastItem; }
                while (IsEmpty())
                {
                    Monitor.Wait(Queue);
                }
                int element = Queue.Dequeue();
                if (element == Producer.LastItem) { _ending = true; }
                if (Queue.Count == _capacity - 1)
                {
                    Monitor.PulseAll(Queue);
                }
                return element;
            }
            finally
            {
                Monitor.Exit(Queue);
            }
        }

    }
}
